@@ -2,11 +2,8 @@
  *
  * raptor_sax2.c - Raptor SAX2 API
  *
- * $Id: raptor_sax2.c,v 1.30 2005/09/20 21:16:28 cmdjb Exp $
- *
- * Copyright (C) 2000-2005, David Beckett http://purl.org/net/dajobe/
- * Institute for Learning and Research Technology http://www.ilrt.bristol.ac.uk/
- * University of Bristol, UK http://www.bristol.ac.uk/
+ * Copyright (C) 2000-2006, David Beckett http://purl.org/net/dajobe/
+ * Copyright (C) 2000-2005, University of Bristol, UK http://www.bristol.ac.uk/
  * 
  * This package is Free Software and part of Redland http://librdf.org/
  * 
@@ -116,16 +113,16 @@ raptor_free_sax2(raptor_sax2 *sax2) {
     raptor_libxml_free(sax2->xc);
     sax2->xc=NULL;
   }
-
-#ifdef RAPTOR_LIBXML_MY_ENTITIES
-  raptor_libxml_libxml_free_entities(rdf_parser);
-#endif
-
 #endif
 
   while( (xml_element=raptor_xml_element_pop(sax2)) )
     raptor_free_xml_element(xml_element);
 
+  raptor_namespaces_clear(&sax2->namespaces);
+
+  if(sax2->base_uri)
+    raptor_free_uri(sax2->base_uri);
+
   RAPTOR_FREE(raptor_sax2, sax2);
 }
 
@@ -187,6 +184,31 @@ raptor_sax2_set_external_entity_ref_handler(raptor_sax2* sax2,
 }
 
 
+/**
+ * raptor_sax2_set_namespace_handler:
+ * @parser: #raptor_sax2 object
+ * @user_data: user data pointer for callback
+ * @handler: new namespace callback function
+ *
+ * Set the XML namespace handler function.
+ *
+ * When a prefix/namespace is seen in an XML parser, call the given
+ * @handler with the prefix string and the #raptor_uri namespace URI.
+ * Either can be NULL for the default prefix or default namespace.
+ *
+ * The handler function does not deal with duplicates so any
+ * namespace may be declared multiple times when a namespace is seen
+ * in different parts of a document.
+ * 
+ **/
+void
+raptor_sax2_set_namespace_handler(raptor_sax2* sax2,
+                                  raptor_namespace_handler handler)
+{
+  sax2->namespace_handler=handler;
+}
+
+
 void
 raptor_sax2_set_locator(raptor_sax2* sax2, raptor_locator* locator)
 {
@@ -220,8 +242,17 @@ raptor_xml_element_push(raptor_sax2 *sax2, raptor_xml_element* element)
 }
 
 
+int
+raptor_xml_element_is_empty(raptor_xml_element* xml_element)
+{
+  return !xml_element->content_cdata_seen &&
+         !xml_element->content_element_seen;
+}
+
+
 const unsigned char*
-raptor_sax2_inscope_xml_language(raptor_sax2 *sax2) {
+raptor_sax2_inscope_xml_language(raptor_sax2 *sax2)
+{
   raptor_xml_element* xml_element;
   
   for(xml_element=sax2->current_element;
@@ -235,7 +266,8 @@ raptor_sax2_inscope_xml_language(raptor_sax2 *sax2) {
 
 
 raptor_uri*
-raptor_sax2_inscope_base_uri(raptor_sax2 *sax2) {
+raptor_sax2_inscope_base_uri(raptor_sax2 *sax2)
+{
   raptor_xml_element *xml_element;
   
   for(xml_element=sax2->current_element; 
@@ -244,32 +276,69 @@ raptor_sax2_inscope_base_uri(raptor_sax2 *sax2) {
     if(xml_element->base_uri)
       return xml_element->base_uri;
     
-  return NULL;
+  return sax2->base_uri;
 }
 
 
 int
-raptor_sax2_get_depth(raptor_sax2 *sax2) {
+raptor_sax2_get_depth(raptor_sax2 *sax2)
+{
   return sax2->depth;
 }
 
 void
-raptor_sax2_inc_depth(raptor_sax2 *sax2) {
+raptor_sax2_inc_depth(raptor_sax2 *sax2)
+{
   sax2->depth++;
 }
 
 void
-raptor_sax2_dec_depth(raptor_sax2 *sax2) {
+raptor_sax2_dec_depth(raptor_sax2 *sax2)
+{
   sax2->depth--;
 }
 
 
+/*
+ * raptor_sax2_simple_error - Error from a sax2 - Internal
+ *
+ * Matches the raptor_simple_message_handler API but calls
+ * the sax2 error_handler
+ */
+static void
+raptor_sax2_simple_error(void* user_data, const char *message, ...)
+{
+  raptor_sax2* sax2=(raptor_sax2*)user_data;
+  va_list arguments;
+
+  va_start(arguments, message);
+
+  if(sax2)
+    raptor_invoke_message_varargs("error",
+                                  sax2->error_handler,
+                                  sax2->error_data,
+                                  sax2->locator,
+                                  message, arguments);
+  
+  va_end(arguments);
+}
+
+
 void
-raptor_sax2_parse_start(raptor_sax2* sax2, raptor_uri *base_uri) {
+raptor_sax2_parse_start(raptor_sax2* sax2, raptor_uri *base_uri)
+{
+  raptor_uri_handler *uri_handler;
+  void *uri_context;
+
   sax2->depth=0;
   sax2->root_element=NULL;
   sax2->current_element=NULL;
 
+  if(base_uri)
+    sax2->base_uri=raptor_uri_copy(base_uri);
+  else
+    sax2->base_uri=NULL;
+
 #ifdef RAPTOR_XML_EXPAT
   if(sax2->xp) {
     XML_ParserFree(sax2->xp);
@@ -291,6 +360,15 @@ raptor_sax2_parse_start(raptor_sax2* sax2, raptor_uri *base_uri) {
     sax2->xc=NULL;
   }
 #endif
+
+  raptor_namespaces_clear(&sax2->namespaces);
+
+  raptor_uri_get_handler(&uri_handler, &uri_context);
+  raptor_namespaces_init(&sax2->namespaces,
+                         uri_handler, uri_context,
+                         raptor_sax2_simple_error, sax2, 
+                         1);
+
 }
 
 
@@ -446,6 +524,54 @@ raptor_sax2_parse_chunk(raptor_sax2* sax2, const unsigned char *buffer,
 }
 
 
+/**
+ * raptor_sax2_set_feature:
+ * @sax2: #raptor_sax2 SAX2 object
+ * @feature: feature to set from enumerated #raptor_feature values
+ * @value: integer feature value (0 or larger)
+ *
+ * Set various SAX2 features.
+ * 
+ * The allowed features are available via raptor_sax2_features_enumerate().
+ *
+ * Return value: non 0 on failure or if the feature is unknown
+ **/
+int
+raptor_sax2_set_feature(raptor_sax2 *sax2, raptor_feature feature, int value)
+{
+  if(value < 0)
+    return -1;
+  
+  switch(feature) {
+    case RAPTOR_FEATURE_NORMALIZE_LANGUAGE:
+      sax2->feature_normalize_language=value;
+      break;
+
+    case RAPTOR_FEATURE_SCANNING:
+    case RAPTOR_FEATURE_ASSUME_IS_RDF:
+    case RAPTOR_FEATURE_ALLOW_NON_NS_ATTRIBUTES:
+    case RAPTOR_FEATURE_ALLOW_OTHER_PARSETYPES:
+    case RAPTOR_FEATURE_ALLOW_BAGID:
+    case RAPTOR_FEATURE_ALLOW_RDF_TYPE_RDF_LIST:
+    case RAPTOR_FEATURE_NON_NFC_FATAL:
+    case RAPTOR_FEATURE_WARN_OTHER_PARSETYPES:
+    case RAPTOR_FEATURE_CHECK_RDF_ID:
+    case RAPTOR_FEATURE_RELATIVE_URIS:
+    case RAPTOR_FEATURE_START_URI:
+    case RAPTOR_FEATURE_WRITER_AUTO_INDENT:
+    case RAPTOR_FEATURE_WRITER_AUTO_EMPTY:
+    case RAPTOR_FEATURE_WRITER_INDENT_WIDTH:
+    case RAPTOR_FEATURE_WRITER_XML_VERSION:
+    case RAPTOR_FEATURE_WRITER_XML_DECLARATION:
+    default:
+      return -1;
+      break;
+  }
+
+  return 0;
+}
+
+
 void
 raptor_sax2_update_document_locator(raptor_sax2* sax2, 
                                     raptor_locator* locator)
@@ -465,7 +591,16 @@ raptor_sax2_start_element(void* user_data, const unsigned char *name,
                           const unsigned char **atts)
 {
   raptor_sax2* sax2=(raptor_sax2*)user_data;
-  
+  raptor_qname* el_name;
+  unsigned char **xml_atts_copy=NULL;
+  size_t xml_atts_size=0;
+  int all_atts_count=0;
+  int ns_attributes_count=0;
+  raptor_qname** named_attrs=NULL;
+  raptor_xml_element* xml_element=NULL;
+  unsigned char *xml_language=NULL;
+  raptor_uri *xml_base=NULL;
+
 #ifdef RAPTOR_XML_EXPAT
 #ifdef EXPAT_UTF8_BOM_CRASH
   sax2->tokens_count++;
@@ -507,8 +642,159 @@ raptor_sax2_start_element(void* user_data, const unsigned char *name,
   }
 #endif
 
+  raptor_sax2_inc_depth(sax2);
+
+  if(atts) {
+    int i;
+
+    /* Save passed in XML attributes pointers so we can 
+     * NULL the pointers when they get handled below (various atts[i]=NULL)
+     */
+    for (i = 0; atts[i]; i++);
+    xml_atts_size=sizeof(unsigned char*) * i;
+    if(xml_atts_size) {
+      xml_atts_copy=(unsigned char**)RAPTOR_MALLOC(cstringpointer,xml_atts_size);
+      memcpy(xml_atts_copy, atts, xml_atts_size);
+    }
+
+    /* XML attributes processing:
+     *   xmlns*   - XML namespaces (Namespaces in XML REC)
+     *     Deleted and used to synthesise namespaces declarations
+     *   xml:lang - XML language (XML REC)
+     *     Deleted and optionally normalised to lowercase
+     *   xml:base - XML Base (XML Base REC)
+     *     Deleted and used to set the in-scope base URI for this XML element
+     */
+    for (i = 0; atts[i]; i+= 2) {
+      all_atts_count++;
+
+      if(strncmp((char*)atts[i], "xml", 3)) {
+        /* count and skip non xml* attributes */
+        ns_attributes_count++;
+        continue;
+      }
+
+      /* synthesise the XML namespace events */
+      if(!memcmp((const char*)atts[i], "xmlns", 5)) {
+        const unsigned char *prefix=atts[i][5] ? &atts[i][6] : NULL;
+        const unsigned char *namespace_name=atts[i+1];
+
+        raptor_namespace* nspace;
+        nspace=raptor_new_namespace(&sax2->namespaces,
+                                    prefix, namespace_name,
+                                    raptor_sax2_get_depth(sax2));
+
+        if(nspace) {
+          raptor_namespaces_start_namespace(&sax2->namespaces, nspace);
+
+          if(sax2->namespace_handler)
+            (*sax2->namespace_handler)(sax2->user_data, nspace);
+        }
+      } else if(!strcmp((char*)atts[i], "xml:lang")) {
+        xml_language=(unsigned char*)RAPTOR_MALLOC(cstring, strlen((char*)atts[i+1])+1);
+        if(!xml_language) {
+          sax2->fatal_error_handler(sax2->fatal_error_data, sax2->locator, "Out of memory");
+          return;
+        }
+
+        /* optionally normalize language to lowercase */
+        if(sax2->feature_normalize_language) {
+          unsigned char *from=(unsigned char*)atts[i+1];
+          unsigned char *to=xml_language;
+          
+          while(*from) {
+            if(isupper(*from))
+              *to++ =tolower(*from++);
+            else
+              *to++ =*from++;
+          }
+          *to='\0';
+        } else
+          strcpy((char*)xml_language, (char*)atts[i+1]);
+      } else if(!strcmp((char*)atts[i], "xml:base")) {
+        raptor_uri* base_uri;
+        raptor_uri* xuri;
+        base_uri=raptor_sax2_inscope_base_uri(sax2);
+        xuri=raptor_new_uri_relative_to_base(base_uri, atts[i+1]);
+        xml_base=raptor_new_uri_for_xmlbase(xuri);
+        raptor_free_uri(xuri);
+      }
+
+      /* delete all xml attributes whether processed above or not */
+      atts[i]=NULL; 
+    }
+  }
+
+
+  /* Create new element structure */
+  el_name=raptor_new_qname(&sax2->namespaces, name, NULL,
+                           raptor_sax2_simple_error, sax2);
+  if(!el_name)
+    return;
+
+  xml_element=raptor_new_xml_element(el_name, xml_language, xml_base);
+  if(!xml_element) {
+    raptor_free_qname(el_name);
+    return;
+  }
+
+  /* Turn string attributes into namespaced-attributes */
+  if(ns_attributes_count) {
+    int i;
+    int offset = 0;
+
+    /* Allocate new array to hold namespaced-attributes */
+    named_attrs=(raptor_qname**)RAPTOR_CALLOC(raptor_qname_array, 
+                                              ns_attributes_count, 
+                                              sizeof(raptor_qname*));
+    if(!named_attrs) {
+      sax2->fatal_error_handler(sax2->fatal_error_data, sax2->locator, "Out of memory");
+      RAPTOR_FREE(raptor_xml_element, xml_element);
+      raptor_free_qname(raptor_xml_element_get_name(xml_element));
+      return;
+    }
+
+    for (i = 0; i < all_atts_count; i++) {
+      raptor_qname* attr;
+
+      /* Skip previously processed attributes */
+      if(!atts[i<<1])
+        continue;
+
+      /* namespace-name[i] stored in named_attrs[i] */
+      attr=raptor_new_qname(&sax2->namespaces,
+                            atts[i<<1], atts[(i<<1)+1],
+                            raptor_sax2_simple_error, sax2);
+      if(!attr) { /* failed - tidy up and return */
+        int j;
+
+        for (j=0; j < i; j++)
+          RAPTOR_FREE(raptor_qname, named_attrs[j]);
+        RAPTOR_FREE(raptor_qname_array, named_attrs);
+        raptor_free_xml_element(xml_element);
+        return;
+      }
+
+      named_attrs[offset++]=attr;
+    }
+  } /* end if ns_attributes_count */
+
+
+  if(named_attrs)
+    raptor_xml_element_set_attributes(xml_element,
+                                      named_attrs, ns_attributes_count);
+
+  raptor_xml_element_push(sax2, xml_element);
+
   if(sax2->start_element_handler)
-    sax2->start_element_handler(sax2->user_data, name, atts);
+    sax2->start_element_handler(sax2->user_data, xml_element);
+
+  if(xml_atts_copy) {
+    /* Restore passed in XML attributes, free the copy */
+    memcpy((void*)atts, xml_atts_copy, xml_atts_size);
+    RAPTOR_FREE(cstringpointer, xml_atts_copy);
+  }
+
 }
 
 
@@ -517,24 +803,45 @@ void
 raptor_sax2_end_element(void* user_data, const unsigned char *name)
 {
   raptor_sax2* sax2=(raptor_sax2*)user_data;
+  raptor_xml_element* xml_element;
+
 #ifdef RAPTOR_XML_EXPAT
 #ifdef EXPAT_UTF8_BOM_CRASH
   sax2->tokens_count++;
 #endif
 #endif
 
-  if(sax2->end_element_handler)
-    sax2->end_element_handler(sax2->user_data, name);
+  xml_element=sax2->current_element;
+  if(xml_element) {
+#ifdef RAPTOR_DEBUG_VERBOSE
+    fprintf(stderr, "\nraptor_rdfxml_end_element_handler: End ns-element: ");
+    raptor_qname_print(stderr, xml_element->name);
+    fputc('\n', stderr);
+#endif
+
+    if(sax2->end_element_handler)
+      sax2->end_element_handler(sax2->user_data, xml_element);
+  }
+  
+  raptor_namespaces_end_for_depth(&sax2->namespaces, 
+                                  raptor_sax2_get_depth(sax2));
+  xml_element=raptor_xml_element_pop(sax2);
+  if(xml_element)
+    raptor_free_xml_element(xml_element);
+
+  raptor_sax2_dec_depth(sax2);
 }
 
 
+
+
 /* characters */
 void
 raptor_sax2_characters(void* user_data, const unsigned char *s, int len)
 {
   raptor_sax2* sax2=(raptor_sax2*)user_data;
   if(sax2->characters_handler)
-    sax2->characters_handler(sax2->user_data, s, len);
+    sax2->characters_handler(sax2->user_data, sax2->current_element, s, len);
 }
 
 
@@ -550,7 +857,7 @@ raptor_sax2_cdata(void* user_data, const unsigned char *s, int len)
 #endif
 
   if(sax2->cdata_handler)
-    sax2->cdata_handler(sax2->user_data, s, len);
+    sax2->cdata_handler(sax2->user_data, sax2->current_element, s, len);
 }
 
 
@@ -560,7 +867,7 @@ raptor_sax2_comment(void* user_data, const unsigned char *s)
 {
   raptor_sax2* sax2=(raptor_sax2*)user_data;
   if(sax2->comment_handler)
-    sax2->comment_handler(sax2->user_data, s);
+    sax2->comment_handler(sax2->user_data, sax2->current_element, s);
 }
 
 
