@@ -1,12 +1,9 @@
 /* -*- Mode: c; c-basic-offset: 2 -*-
  *
- * raptor_serialize_rss.c - Raptor RSS 1.0 serializer
+ * raptor_serialize_rss.c - Raptor RSS 1.0 and Atom 1.0 serializers
  *
- * $Id: raptor_serialize_rss.c,v 1.95 2005/11/03 07:06:28 dajobe Exp $
- *
- * Copyright (C) 2003-2005, David Beckett http://purl.org/net/dajobe/
- * Institute for Learning and Research Technology http://www.ilrt.bristol.ac.uk/
- * University of Bristol, UK http://www.bristol.ac.uk/
+ * Copyright (C) 2003-2006, David Beckett http://purl.org/net/dajobe/
+ * Copyright (C) 2003-2005, University of Bristol, UK http://www.bristol.ac.uk/
  * 
  * Contributions:
  *   Copyright (C) 2004-2005, Suzan Foster <su@islief.nl>
@@ -200,7 +197,7 @@ raptor_rss10_move_statements(raptor_rss10_serializer_context *rss_serializer,
           if(s->object_type == RAPTOR_IDENTIFIER_TYPE_RESOURCE)
             field->uri=(raptor_uri*)s->object;
           else
-            field->value=(char*)s->object;
+            field->value=(unsigned char*)s->object;
           s->object=NULL;
 
           if(is_atom) { 
@@ -324,7 +321,7 @@ raptor_rss10_store_statement(raptor_rss10_serializer_context *rss_serializer,
           field->uri=(raptor_uri*)s->object;
           s->object=NULL;
         } else {
-          field->value=(char*)s->object;
+          field->value=(unsigned char*)s->object;
           s->object=NULL;
         }
 
@@ -782,10 +779,12 @@ raptor_rss10_emit_item(raptor_serializer* serializer,
       } else {
         /* not a URI, must be a literal */
         raptor_xml_writer_start_element(xml_writer, predicate);
-        if(f == RAPTOR_RSS_FIELD_CONTENT_ENCODED) {
+        if(!is_atom && f == RAPTOR_RSS_FIELD_CONTENT_ENCODED) {
           raptor_xml_writer_raw_counted(xml_writer, (const unsigned char*)"<![CDATA[", 9);
           raptor_xml_writer_raw(xml_writer, (const unsigned char*)field->value);
           raptor_xml_writer_raw_counted(xml_writer, (const unsigned char*)"]]>", 3);
+        } else if(is_atom && f == RAPTOR_RSS_FIELD_ATOM_SUMMARY) {
+          raptor_xml_writer_raw(xml_writer, (const unsigned char*)field->value);
         } else
           raptor_xml_writer_cdata(xml_writer, (const unsigned char*)field->value);
         raptor_xml_writer_end_element(xml_writer, predicate);
@@ -916,9 +915,6 @@ raptor_rss10_serialize_end(raptor_serializer* serializer) {
                                    1);
   rss_serializer->xml_writer=xml_writer;
 
-  raptor_xml_writer_raw(xml_writer,
-                        (const unsigned char*)"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
-
   raptor_rss10_build_xml_names(serializer);
 
   raptor_xml_writer_start_element(xml_writer, rss_serializer->root_element);
